"""
Collections of bootstrap methods and classes.

How to use
----------

test

"""

__author__ = 'Felix Kleinert, Lukas Leufen'
__date__ = '2020-02-07'


import logging
import os
import re
from typing import List, Union, Pattern, Tuple

import dask.array as da
import keras
import numpy as np
import xarray as xr

from mlair import helpers
from mlair.data_handling.data_generator import DataGenerator


class BootStrapGenerator(keras.utils.Sequence):
    """
    Generator that returns bootstrapped history objects for given boot index while iteration.

    generator for bootstraps as keras sequence inheritance. Initialise with number of boots, the original history, the
    shuffled data, all used variables and the current shuffled variable. While iterating over this generator, it returns
    the bootstrapped history for given boot index (this is the iterator index) in the same format like the original
    history ready to use. Note, that in some cases some samples can contain nan values (in these cases the entire data
    row is null, not only single entries).
    """

    def __init__(self, number_of_boots: int, history: xr.DataArray, shuffled: xr.DataArray, variables: List[str],
                 shuffled_variable: str):
        """
        Set up the generator.

        :param number_of_boots: number of bootstrap realisations
        :param history: original history (the ground truth)
        :param shuffled: the shuffled history
        :param variables: list with all variables of interest
        :param shuffled_variable: name of the variable that shall be bootstrapped
        """
        self.number_of_boots = number_of_boots
        self.variables = variables
        self.history_orig = history
        self.history = history.sel(variables=helpers.remove_items(self.variables, shuffled_variable))
        self.shuffled = shuffled.sel(variables=shuffled_variable)

    def __len__(self) -> int:
        """
        Return number of bootstraps.

        :return: number of bootstraps
        """
        return self.number_of_boots

    def __getitem__(self, index: int) -> xr.DataArray:
        """
        Return bootstrapped history for given bootstrap index in same index structure like the original history object.

        :param index: boot index e [0, nboots-1]
        :return: bootstrapped history ready to use
        """
        logging.debug(f"boot: {index}")
        boot_hist = self.history.copy()
        boot_hist = boot_hist.combine_first(self.__get_shuffled(index))
        return boot_hist.reindex_like(self.history_orig)

    def __get_shuffled(self, index: int) -> xr.DataArray:
        """
        Return shuffled data for given boot index from shuffled attribute.

        :param index: boot index e [0, nboots-1]
        :return: shuffled data
        """
        shuffled_var = self.shuffled.sel(boots=index).expand_dims("variables").drop("boots")
        return shuffled_var.transpose("datetime", "window", "Stations", "variables")


class CreateShuffledData:
    """
    Verify and create shuffled data for all data contained in given data generator class.

    Starts automatically on initialisation, no further calls are required. Check and new creations are all performed
    inside bootstrap_path.
    """

    def __init__(self, data: DataGenerator, number_of_bootstraps: int, bootstrap_path: str):
        """
        Shuffled data is automatically created in initialisation.

        :param data: data to shuffle
        :param number_of_bootstraps:
        :param bootstrap_path: Path to find and store the bootstraps
        """
        self.data = data
        self.number_of_bootstraps = number_of_bootstraps
        self.bootstrap_path = bootstrap_path
        self.create_shuffled_data()

    def create_shuffled_data(self) -> None:
        """
        Create shuffled data.

        Use original test data, add dimension 'boots' with length number of bootstraps and insert randomly selected
        variables. If there is a suitable local file for requested window size and number of bootstraps, no additional
        file will be created inside this function.
        """
        logging.info("create / check shuffled bootstrap data")
        variables_str = '_'.join(sorted(self.data.variables))
        window = self.data.window_history_size
        for station in self.data.stations:
            valid, nboot = self.valid_bootstrap_file(station, variables_str, window)
            if not valid:
                logging.info(f'create bootstap data for {station}')
                hist = self.data.get_data_generator(station).get_transposed_history()
                file_path = self._set_file_path(station, variables_str, window, nboot)
                hist = hist.expand_dims({'boots': range(nboot)}, axis=-1)
                shuffled_variable = []
                chunks = (100, *hist.shape[1:3], hist.shape[-1])
                for i, var in enumerate(hist.coords['variables']):
                    single_variable = hist.sel(variables=var).values
                    shuffled_variable.append(self.shuffle(single_variable, chunks=chunks))
                shuffled_variable_da = da.stack(shuffled_variable, axis=-2).rechunk("auto")
                shuffled_data = xr.DataArray(shuffled_variable_da, coords=hist.coords, dims=hist.dims)
                shuffled_data.to_netcdf(file_path)

    def _set_file_path(self, station: str, variables: str, window: int, nboots: int) -> str:
        """
        Set file name.

        Set file name following naming convention <station>_<var1>_<var2>_..._hist<window>_nboots<nboots>_shuffled.nc
        and create joined path using bootstrap_path attribute set on initialisation.

        :param station: station name
        :param variables: variables already preprocessed as single string with all variables seperated by underscore
        :param window: window length
        :param nboots: number of boots
        :return: full file path
        """
        file_name = f"{station}_{variables}_hist{window}_nboots{nboots}_shuffled.nc"
        return os.path.join(self.bootstrap_path, file_name)

    def valid_bootstrap_file(self, station: str, variables: str, window: int) -> [bool, Union[None, int]]:
        """
        Compare local bootstrap file with given settings for station, variables, window and number of bootstraps.

        If a match was found, this method returns a tuple (True, None). In any other case, it returns (False,
        max_nboot), where max_nboot is the highest boot number found in the local storage. A match is defined so that
        the window length is ge than given window size form args and the number of boots is also ge than the given
        number of boots from this class. Furthermore, this functions deletes local files, if the match the station
        pattern but don't fit the window and bootstrap condition. This is performed, because it is assumed, that the
        corresponding file will be created with a longer or at the least same window size and numbers of bootstraps.

        :param station: name of the station to validate
        :param variables: all variables already merged in single string seperated by underscore
        :param window: required window size
        :return: tuple containing information if valid file was found first and second the number of boots that needs to
            be used for the new boot creation (this is only relevant, if no valid file was found - otherwise the return
            statement is anyway None).
        """
        regex = re.compile(rf"{station}_{variables}_hist(\d+)_nboots(\d+)_shuffled")
        max_nboot = self.number_of_bootstraps
        for file in os.listdir(self.bootstrap_path):
            match = regex.match(file)
            if match:
                window_file = int(match.group(1))
                nboot_file = int(match.group(2))
                max_nboot = max([max_nboot, nboot_file])
                if (window_file >= window) and (nboot_file >= self.number_of_bootstraps):
                    return True, None
                else:
                    os.remove(os.path.join(self.bootstrap_path, file))
        return False, max_nboot

    @staticmethod
    def shuffle(data: da.array, chunks: Tuple) -> da.core.Array:
        """
        Shuffle randomly from given data (draw elements with replacement).

        :param data: data to shuffle
        :param chunks: chunk size for dask
        :return: shuffled data as dask core array (not computed yet)
        """
        size = data.shape
        return da.random.choice(data.reshape(-1, ), size=size, chunks=chunks)


class BootStraps:
    """
    Main class to perform bootstrap operations.

    This class requires a DataGenerator object and a path, where to find and store all data related to the bootstrap
    operation. In initialisation, this class will automatically call the class CreateShuffleData to set up the shuffled
    data sets. How to use BootStraps:

    * call .get_generator(<station>, <variable>) to get a generator for given station and variable combination that \
        iterates over all bootstrap realisations (as keras sequence)
    * call .get_labels(<station>) to get the measured observations in the same format as bootstrap predictions
    * call .get_bootstrap_predictions(<station>, <variable>) to get the bootstrapped predictions
    * call .get_orig_prediction(<station>) to get the non-bootstrapped predictions (referred as original predictions)
    """

    def __init__(self, data: DataGenerator, bootstrap_path: str, number_of_bootstraps: int = 10):
        """
        Automatically check and create (if needed) shuffled data on initialisation.

        :param data: a data generator object to get data / history
        :param bootstrap_path: path to find and store the bootstrap data
        :param number_of_bootstraps: the number of bootstrap realisations
        """
        self.data = data
        self.number_of_bootstraps = number_of_bootstraps
        self.bootstrap_path = bootstrap_path
        CreateShuffledData(data, number_of_bootstraps, bootstrap_path)

    @property
    def stations(self) -> List[str]:
        """
        Station property inherits directly from data generator object.

        :return: list with all stations
        """
        return self.data.stations

    @property
    def variables(self) -> List[str]:
        """
        Variables property inherits directly from data generator object.

        :return: list with all variables
        """
        return self.data.variables

    @property
    def window_history_size(self) -> int:
        """
        Window history size property inherits directly from data generator object.

        :return: the window history size
        """
        return self.data.window_history_size

    def get_generator(self, station: str, variable: str) -> BootStrapGenerator:
        """
        Return the actual generator to use for the bootstrap evaluation.

        The generator requires information on station and bootstrapped variable. There is only a loop on the bootstrap
        realisation and not on stations or variables.

        :param station: name of the station
        :param variable: name of the variable to bootstrap
        :return: BootStrapGenerator class ready to use.
        """
        hist, _ = self.data[station]
        shuffled_data = self._load_shuffled_data(station, self.variables).reindex_like(hist)
        return BootStrapGenerator(self.number_of_bootstraps, hist, shuffled_data, self.variables, variable)

    def get_labels(self, station: str) -> np.ndarray:
        """
        Repeat labels for given key by the number of boots and returns as single array.

        :param station: name of station
        :return: repeated labels as single array
        """
        labels = self.data[station][1]
        return np.tile(labels.data, (self.number_of_bootstraps, 1))

    def get_orig_prediction(self, path: str, file_name: str, prediction_name: str = "CNN") -> np.ndarray:
        """
        Repeat predictions from given file(_name) in path by the number of boots.

        :param path: path to file
        :param file_name: file name
        :param prediction_name: name of the prediction to select from loaded file (default CNN)
        :return: repeated predictions
        """
        file = os.path.join(path, file_name)
        prediction = xr.open_dataarray(file).sel(type=prediction_name).squeeze()
        vals = np.tile(prediction.data, (self.number_of_bootstraps, 1))
        return vals[~np.isnan(vals).any(axis=1), :]

    def _load_shuffled_data(self, station: str, variables: List[str]) -> xr.DataArray:
        """
        Load shuffled data from bootstrap path.

        Data is stored as '<station>_<var1>_<var2>_..._hist<histsize>_nboots<nboots>_shuffled.nc', e.g.
        'DEBW107_cloudcover_no_no2_temp_u_v_hist13_nboots20_shuffled.nc'

        :param station: name of station
        :param variables: list of variables
        :return: shuffled data as xarray
        """
        file_name = self._get_shuffled_data_file(station, variables)
        shuffled_data = xr.open_dataarray(file_name, chunks=100)
        return shuffled_data

    def _get_shuffled_data_file(self, station: str, variables: List[str]) -> str:
        """
        Look for data file using regular expressions and returns found file or raise FileNotFoundError.

        :param station: name of station
        :param variables: name of variables
        :return: found file with complete path
        """
        files = os.listdir(self.bootstrap_path)
        regex = self._create_file_regex(station, variables)
        file = self._filter_files(regex, files, self.window_history_size, self.number_of_bootstraps)
        if file:
            return os.path.join(self.bootstrap_path, file)
        else:
            raise FileNotFoundError(f"Could not find a file to match pattern {regex}")

    @staticmethod
    def _create_file_regex(station: str, variables: List[str]) -> Pattern:
        """
        Create regex for given station and variables.

        With this regex, it is possible to look for shuffled data with pattern:
        `<station>(_<var>)*_hist(<hist>)_nboots(<nboots>)_shuffled.nc`

        :param station: station name to use as prefix
        :param variables: variables to add after station
        :return: compiled regular expression
        """
        var_regex = "".join([rf"(_\w+)*_{v}(_\w+)*" for v in sorted(variables)])
        regex = re.compile(rf"{station}{var_regex}_hist(\d+)_nboots(\d+)_shuffled\.nc")
        return regex

    @staticmethod
    def _filter_files(regex: Pattern, files: List[str], window: int, nboot: int) -> Union[str, None]:
        """
        Filter list of files by regex.

        Regex has to be structured to match the following string structure
        `<station>(_<var>)*_hist(<hist>)_nboots(<nboots>)_shuffled.nc`. Hist and nboots values have to be included as
        group. All matches are compared to given window and nboot parameters. A valid file must have the same value (or
        larger) than these parameters and contain all variables.

        :param regex: compiled regular expression pattern following the style from method description
        :param files: list of file names to filter
        :param window: minimum length of window to look for
        :param nboot: minimal number of boots to search
        :return: matching file name or None, if no valid file was found
        """
        for f in files:
            match = regex.match(f)
            if match:
                last = match.lastindex
                if (int(match.group(last - 1)) >= window) and (int(match.group(last)) >= nboot):
                    return f


if __name__ == "__main__":

    from mlair.run_modules.experiment_setup import ExperimentSetup
    from mlair.run_modules.run_environment import RunEnvironment
    from mlair.run_modules.pre_processing import PreProcessing

    formatter = '%(asctime)s - %(levelname)s: %(message)s  [%(filename)s:%(funcName)s:%(lineno)s]'
    logging.basicConfig(format=formatter, level=logging.INFO)

    with RunEnvironment() as run_env:
        ExperimentSetup(stations=['DEBW107', 'DEBY081', 'DEBW013'],
                        station_type='background', trainable=True, window_history_size=9)
        PreProcessing()

        data = run_env.data_store.get("generator", "general.test")
        path = run_env.data_store.get("bootstrap_path", "general")
        number_bootstraps = 10

        boots = BootStraps(data, path, number_bootstraps)
        for b in boots.boot_strap_generator():
            a, c = b
        logging.info(f"len is {len(boots.get_boot_strap_meta())}")
